package org.eclipse.jdt.internal.compiler.ast;

package .var.folders._j.vfbv187d0bb8b8w5spzgg8180000gn.T;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import org.eclipse.jdt.core.compiler.CharOperation;
import org.eclipse.jdt.internal.compiler.CompilationResult;
import org.eclipse.jdt.internal.compiler.ast.ASTNode;
import org.eclipse.jdt.internal.compiler.ast.AllocationExpression;
import org.eclipse.jdt.internal.compiler.ast.Assignment;
import org.eclipse.jdt.internal.compiler.ast.CastExpression;
import org.eclipse.jdt.internal.compiler.ast.ConditionalExpression;
import org.eclipse.jdt.internal.compiler.ast.Expression;
import org.eclipse.jdt.internal.compiler.ast.LocalDeclaration;
import org.eclipse.jdt.internal.compiler.ast.MessageSend;
import org.eclipse.jdt.internal.compiler.ast.QualifiedNameReference;
import org.eclipse.jdt.internal.compiler.ast.SingleNameReference;
import org.eclipse.jdt.internal.compiler.ast.SingleTypeReference;
import org.eclipse.jdt.internal.compiler.ast.TypeReference;
import org.eclipse.jdt.internal.compiler.codegen.CodeStream;
import org.eclipse.jdt.internal.compiler.flow.FinallyFlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowContext;
import org.eclipse.jdt.internal.compiler.flow.FlowInfo;
import org.eclipse.jdt.internal.compiler.impl.Constant;
import org.eclipse.jdt.internal.compiler.impl.ReferenceContext;
import org.eclipse.jdt.internal.compiler.lookup.BlockScope;
import org.eclipse.jdt.internal.compiler.lookup.LocalVariableBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodBinding;
import org.eclipse.jdt.internal.compiler.lookup.MethodScope;
import org.eclipse.jdt.internal.compiler.lookup.ReferenceBinding;
import org.eclipse.jdt.internal.compiler.lookup.Scope;
import org.eclipse.jdt.internal.compiler.lookup.TypeBinding;
import org.eclipse.jdt.internal.compiler.lookup.TypeConstants;
import org.eclipse.jdt.internal.compiler.problem.ProblemReporter;
import org.eclipse.jdt.internal.compiler.util.Util;

public class FakedTrackingVariable extends LocalDeclaration {
  private static final char[] UNASSIGNED_CLOSEABLE_NAME = "<unassigned Closeable value>".toCharArray();
  
  private static final char[] UNASSIGNED_CLOSEABLE_NAME_TEMPLATE = "<unassigned Closeable value from line {0}>".toCharArray();
  
  private static final char[] TEMPLATE_ARGUMENT = "{0}".toCharArray();
  
  private static final int CLOSE_SEEN = 1;
  
  private static final int SHARED_WITH_OUTSIDE = 2;
  
  private static final int OWNED_BY_OUTSIDE = 4;
  
  private static final int CLOSED_IN_NESTED_METHOD = 8;
  
  private static final int REPORTED_EXPLICIT_CLOSE = 16;
  
  private static final int REPORTED_POTENTIAL_LEAK = 32;
  
  private static final int REPORTED_DEFINITIVE_LEAK = 64;
  
  public static boolean TEST_372319 = false;
  
  private int globalClosingState = 0;
  
  public LocalVariableBinding originalBinding;
  
  public org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable innerTracker;
  
  public org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable outerTracker;
  
  MethodScope methodScope;
  
  private HashMap recordedLocations;
  
  private ASTNode currentAssignment;
  
  private FlowContext tryContext;
  
  public FakedTrackingVariable(LocalVariableBinding original, ASTNode location, FlowInfo flowInfo, FlowContext flowContext, int nullStatus) {
    super(original.name, location.sourceStart, location.sourceEnd);
    this.type = (TypeReference)new SingleTypeReference(
        TypeConstants.OBJECT, (
        this.sourceStart << 32L) + this.sourceEnd);
    this.methodScope = original.declaringScope.methodScope();
    this.originalBinding = original;
    while (flowContext != null) {
      if (flowContext instanceof FinallyFlowContext) {
        this.tryContext = ((FinallyFlowContext)flowContext).tryContext;
        break;
      } 
      flowContext = flowContext.parent;
    } 
    resolve(original.declaringScope);
    if (nullStatus != 0)
      flowInfo.markNullStatus(this.binding, nullStatus); 
  }
  
  private FakedTrackingVariable(BlockScope scope, ASTNode location, FlowInfo flowInfo, int nullStatus) {
    super(UNASSIGNED_CLOSEABLE_NAME, location.sourceStart, location.sourceEnd);
    this.type = (TypeReference)new SingleTypeReference(
        TypeConstants.OBJECT, (
        this.sourceStart << 32L) + this.sourceEnd);
    this.methodScope = scope.methodScope();
    this.originalBinding = null;
    resolve(scope);
    if (nullStatus != 0)
      flowInfo.markNullStatus(this.binding, nullStatus); 
  }
  
  public void generateCode(BlockScope currentScope, CodeStream codeStream) {}
  
  public void resolve(BlockScope scope) {
    this.binding = new LocalVariableBinding(
        this.name, 
        (TypeBinding)scope.getJavaLangObject(), 
        0, 
        false);
    this.binding.closeTracker = this;
    this.binding.declaringScope = scope;
    this.binding.setConstant(Constant.NotAConstant);
    this.binding.useFlag = 1;
    this.binding.id = scope.registerTrackingVariable(this);
  }
  
  public static org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable getCloseTrackingVariable(Expression expression, FlowInfo flowInfo, FlowContext flowContext) {
    while (true) {
      while (expression instanceof CastExpression)
        expression = ((CastExpression)expression).expression; 
      if (expression instanceof Assignment) {
        expression = ((Assignment)expression).expression;
        continue;
      } 
      break;
    } 
    if (expression instanceof ConditionalExpression) {
      org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable falseTrackingVariable = getCloseTrackingVariable(((ConditionalExpression)expression).valueIfFalse, flowInfo, flowContext);
      if (falseTrackingVariable != null)
        return falseTrackingVariable; 
      return getCloseTrackingVariable(((ConditionalExpression)expression).valueIfTrue, flowInfo, flowContext);
    } 
    if (expression instanceof SingleNameReference) {
      SingleNameReference name = (SingleNameReference)expression;
      if (name.binding instanceof LocalVariableBinding) {
        LocalVariableBinding local = (LocalVariableBinding)name.binding;
        if (local.closeTracker != null)
          return local.closeTracker; 
        if (!isAnyCloseable(expression.resolvedType))
          return null; 
        if ((local.tagBits & 0x2000L) != 0L)
          return null; 
        LocalDeclaration localDeclaration = local.declaration;
        local.closeTracker = new org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable(local, (ASTNode)localDeclaration, flowInfo, flowContext, 1);
        if (local.isParameter())
          local.closeTracker.globalClosingState |= 0x4; 
        return local.closeTracker;
      } 
    } else if (expression instanceof AllocationExpression) {
      return ((AllocationExpression)expression).closeTracker;
    } 
    return null;
  }
  
  public static void preConnectTrackerAcrossAssignment(ASTNode location, LocalVariableBinding local, Expression rhs, FlowInfo flowInfo) {
    org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable closeTracker = null;
    if (containsAllocation((ASTNode)rhs)) {
      closeTracker = local.closeTracker;
      if (closeTracker == null && 
        rhs.resolvedType != TypeBinding.NULL) {
        closeTracker = new org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable(local, location, flowInfo, null, 1);
        if (local.isParameter())
          closeTracker.globalClosingState |= 0x4; 
      } 
      if (closeTracker != null) {
        closeTracker.currentAssignment = location;
        preConnectTrackerAcrossAssignment(location, local, flowInfo, closeTracker, rhs);
      } 
    } 
  }
  
  private static boolean containsAllocation(ASTNode location) {
    if (location instanceof AllocationExpression)
      return true; 
    if (location instanceof ConditionalExpression) {
      ConditionalExpression conditional = (ConditionalExpression)location;
      return !(!containsAllocation((ASTNode)conditional.valueIfTrue) && !containsAllocation((ASTNode)conditional.valueIfFalse));
    } 
    if (location instanceof CastExpression)
      return containsAllocation((ASTNode)((CastExpression)location).expression); 
    return false;
  }
  
  private static void preConnectTrackerAcrossAssignment(ASTNode location, LocalVariableBinding local, FlowInfo flowInfo, org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable closeTracker, Expression expression) {
    if (expression instanceof AllocationExpression) {
      preConnectTrackerAcrossAssignment(location, local, flowInfo, (AllocationExpression)expression, closeTracker);
    } else if (expression instanceof ConditionalExpression) {
      preConnectTrackerAcrossAssignment(location, local, flowInfo, (ConditionalExpression)expression, closeTracker);
    } else if (expression instanceof CastExpression) {
      preConnectTrackerAcrossAssignment(location, local, ((CastExpression)expression).expression, flowInfo);
    } 
  }
  
  private static void preConnectTrackerAcrossAssignment(ASTNode location, LocalVariableBinding local, FlowInfo flowInfo, ConditionalExpression conditional, org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable closeTracker) {
    preConnectTrackerAcrossAssignment(location, local, flowInfo, closeTracker, conditional.valueIfFalse);
    preConnectTrackerAcrossAssignment(location, local, flowInfo, closeTracker, conditional.valueIfTrue);
  }
  
  private static void preConnectTrackerAcrossAssignment(ASTNode location, LocalVariableBinding local, FlowInfo flowInfo, AllocationExpression allocationExpression, org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable closeTracker) {
    allocationExpression.closeTracker = closeTracker;
    if (allocationExpression.arguments != null && allocationExpression.arguments.length > 0)
      preConnectTrackerAcrossAssignment(location, local, allocationExpression.arguments[0], flowInfo); 
  }
  
  public static void analyseCloseableAllocation(BlockScope scope, FlowInfo flowInfo, AllocationExpression allocation) {
    if (((ReferenceBinding)allocation.resolvedType).hasTypeBit(8)) {
      if (allocation.closeTracker != null) {
        allocation.closeTracker.withdraw();
        allocation.closeTracker = null;
      } 
    } else if (((ReferenceBinding)allocation.resolvedType).hasTypeBit(4)) {
      boolean isWrapper = true;
      if (allocation.arguments != null && allocation.arguments.length > 0) {
        org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable innerTracker = findCloseTracker(scope, flowInfo, allocation.arguments[0]);
        if (innerTracker != null) {
          org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable currentInner = innerTracker;
          while (true) {
            if (currentInner == allocation.closeTracker)
              return; 
            currentInner = currentInner.innerTracker;
            if (currentInner == null) {
              int newStatus = 2;
              if (allocation.closeTracker == null) {
                allocation.closeTracker = new org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable(scope, (ASTNode)allocation, flowInfo, 1);
              } else if (scope.finallyInfo != null) {
                int finallyStatus = scope.finallyInfo.nullStatus(allocation.closeTracker.binding);
                if (finallyStatus != 1)
                  newStatus = finallyStatus; 
              } 
              if (allocation.closeTracker.innerTracker != null)
                innerTracker = pickMoreUnsafe(allocation.closeTracker.innerTracker, innerTracker, scope, flowInfo); 
              allocation.closeTracker.innerTracker = innerTracker;
              innerTracker.outerTracker = allocation.closeTracker;
              flowInfo.markNullStatus(allocation.closeTracker.binding, newStatus);
              if (newStatus != 2) {
                org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable currentTracker = innerTracker;
                while (currentTracker != null) {
                  flowInfo.markNullStatus(currentTracker.binding, newStatus);
                  currentTracker.globalClosingState |= allocation.closeTracker.globalClosingState;
                  currentTracker = currentTracker.innerTracker;
                } 
              } 
              return;
            } 
          } 
        } 
        if (!isAnyCloseable((allocation.arguments[0]).resolvedType))
          isWrapper = false; 
      } else {
        isWrapper = false;
      } 
      if (isWrapper) {
        if (allocation.closeTracker != null) {
          allocation.closeTracker.withdraw();
          allocation.closeTracker = null;
        } 
      } else {
        handleRegularResource(scope, flowInfo, allocation);
      } 
    } else {
      handleRegularResource(scope, flowInfo, allocation);
    } 
  }
  
  private static org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable pickMoreUnsafe(org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable tracker1, org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable tracker2, BlockScope scope, FlowInfo info) {
    int status1 = info.nullStatus(tracker1.binding);
    int status2 = info.nullStatus(tracker2.binding);
    if (status1 == 2 || status2 == 4)
      return pick(tracker1, tracker2, scope); 
    if (status1 == 4 || status2 == 2)
      return pick(tracker2, tracker1, scope); 
    if ((status1 & 0x10) != 0)
      return pick(tracker1, tracker2, scope); 
    if ((status2 & 0x10) != 0)
      return pick(tracker2, tracker1, scope); 
    return pick(tracker1, tracker2, scope);
  }
  
  private static org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable pick(org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable tracker1, org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable tracker2, BlockScope scope) {
    tracker2.withdraw();
    return tracker1;
  }
  
  private static void handleRegularResource(BlockScope scope, FlowInfo flowInfo, AllocationExpression allocation) {
    org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable presetTracker = allocation.closeTracker;
    if (presetTracker != null && presetTracker.originalBinding != null) {
      int closeStatus = flowInfo.nullStatus(presetTracker.binding);
      if (closeStatus != 4 && 
        closeStatus != 1 && 
        !flowInfo.isDefinitelyNull(presetTracker.originalBinding) && 
        !(presetTracker.currentAssignment instanceof LocalDeclaration))
        allocation.closeTracker.recordErrorLocation(presetTracker.currentAssignment, closeStatus); 
    } else {
      allocation.closeTracker = new org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable(scope, (ASTNode)allocation, flowInfo, 1);
    } 
    flowInfo.markAsDefinitelyNull(allocation.closeTracker.binding);
  }
  
  private static org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable findCloseTracker(BlockScope scope, FlowInfo flowInfo, Expression arg) {
    while (arg instanceof Assignment) {
      Assignment assign = (Assignment)arg;
      LocalVariableBinding innerLocal = assign.localVariableBinding();
      if (innerLocal != null)
        return innerLocal.closeTracker; 
      arg = assign.expression;
    } 
    if (arg instanceof SingleNameReference) {
      LocalVariableBinding local = arg.localVariableBinding();
      if (local != null)
        return local.closeTracker; 
    } else if (arg instanceof AllocationExpression) {
      return ((AllocationExpression)arg).closeTracker;
    } 
    return null;
  }
  
  public static void handleResourceAssignment(BlockScope scope, FlowInfo upstreamInfo, FlowInfo flowInfo, FlowContext flowContext, ASTNode location, Expression rhs, LocalVariableBinding local) {
    // Byte code:
    //   0: aconst_null
    //   1: astore #7
    //   3: aconst_null
    //   4: astore #8
    //   6: aload #6
    //   8: getfield closeTracker : Lorg/eclipse/jdt/internal/compiler/ast/FakedTrackingVariable;
    //   11: ifnull -> 45
    //   14: aload #6
    //   16: getfield closeTracker : Lorg/eclipse/jdt/internal/compiler/ast/FakedTrackingVariable;
    //   19: astore #7
    //   21: aload_1
    //   22: aload #6
    //   24: invokevirtual nullStatus : (Lorg/eclipse/jdt/internal/compiler/lookup/LocalVariableBinding;)I
    //   27: istore #9
    //   29: iload #9
    //   31: iconst_2
    //   32: if_icmpeq -> 45
    //   35: iload #9
    //   37: iconst_1
    //   38: if_icmpeq -> 45
    //   41: aload #7
    //   43: astore #8
    //   45: aload #5
    //   47: getfield resolvedType : Lorg/eclipse/jdt/internal/compiler/lookup/TypeBinding;
    //   50: getstatic org/eclipse/jdt/internal/compiler/lookup/TypeBinding.NULL : Lorg/eclipse/jdt/internal/compiler/lookup/NullTypeBinding;
    //   53: if_acmpeq -> 337
    //   56: aload #5
    //   58: aload_2
    //   59: aload_3
    //   60: invokestatic getCloseTrackingVariable : (Lorg/eclipse/jdt/internal/compiler/ast/Expression;Lorg/eclipse/jdt/internal/compiler/flow/FlowInfo;Lorg/eclipse/jdt/internal/compiler/flow/FlowContext;)Lorg/eclipse/jdt/internal/compiler/ast/FakedTrackingVariable;
    //   63: astore #9
    //   65: aload #9
    //   67: ifnull -> 194
    //   70: aload #6
    //   72: getfield closeTracker : Lorg/eclipse/jdt/internal/compiler/ast/FakedTrackingVariable;
    //   75: ifnonnull -> 118
    //   78: aload #9
    //   80: getfield originalBinding : Lorg/eclipse/jdt/internal/compiler/lookup/LocalVariableBinding;
    //   83: ifnull -> 93
    //   86: aload #6
    //   88: aload #9
    //   90: putfield closeTracker : Lorg/eclipse/jdt/internal/compiler/ast/FakedTrackingVariable;
    //   93: aload #9
    //   95: getfield currentAssignment : Lorg/eclipse/jdt/internal/compiler/ast/ASTNode;
    //   98: aload #4
    //   100: if_acmpne -> 337
    //   103: aload #9
    //   105: dup
    //   106: getfield globalClosingState : I
    //   109: bipush #-7
    //   111: iand
    //   112: putfield globalClosingState : I
    //   115: goto -> 337
    //   118: aload #5
    //   120: instanceof org/eclipse/jdt/internal/compiler/ast/AllocationExpression
    //   123: ifne -> 134
    //   126: aload #5
    //   128: instanceof org/eclipse/jdt/internal/compiler/ast/ConditionalExpression
    //   131: ifeq -> 184
    //   134: aload #9
    //   136: aload #8
    //   138: if_acmpne -> 142
    //   141: return
    //   142: aload #6
    //   144: getfield closeTracker : Lorg/eclipse/jdt/internal/compiler/ast/FakedTrackingVariable;
    //   147: aload #9
    //   149: if_acmpne -> 184
    //   152: aload #9
    //   154: getfield globalClosingState : I
    //   157: iconst_4
    //   158: iand
    //   159: ifeq -> 184
    //   162: aload #6
    //   164: new org/eclipse/jdt/internal/compiler/ast/FakedTrackingVariable
    //   167: dup
    //   168: aload #6
    //   170: aload #4
    //   172: aload_2
    //   173: aload_3
    //   174: iconst_2
    //   175: invokespecial <init> : (Lorg/eclipse/jdt/internal/compiler/lookup/LocalVariableBinding;Lorg/eclipse/jdt/internal/compiler/ast/ASTNode;Lorg/eclipse/jdt/internal/compiler/flow/FlowInfo;Lorg/eclipse/jdt/internal/compiler/flow/FlowContext;I)V
    //   178: putfield closeTracker : Lorg/eclipse/jdt/internal/compiler/ast/FakedTrackingVariable;
    //   181: goto -> 337
    //   184: aload #6
    //   186: aload #9
    //   188: putfield closeTracker : Lorg/eclipse/jdt/internal/compiler/ast/FakedTrackingVariable;
    //   191: goto -> 337
    //   194: aload #7
    //   196: ifnull -> 291
    //   199: aload_3
    //   200: astore #10
    //   202: aload #7
    //   204: getfield tryContext : Lorg/eclipse/jdt/internal/compiler/flow/FlowContext;
    //   207: ifnull -> 238
    //   210: goto -> 233
    //   213: aload #7
    //   215: getfield tryContext : Lorg/eclipse/jdt/internal/compiler/flow/FlowContext;
    //   218: aload #10
    //   220: if_acmpne -> 226
    //   223: goto -> 337
    //   226: aload #10
    //   228: getfield parent : Lorg/eclipse/jdt/internal/compiler/flow/FlowContext;
    //   231: astore #10
    //   233: aload #10
    //   235: ifnonnull -> 213
    //   238: aload #7
    //   240: getfield globalClosingState : I
    //   243: bipush #6
    //   245: iand
    //   246: ifne -> 270
    //   249: aload_2
    //   250: aload #7
    //   252: getfield binding : Lorg/eclipse/jdt/internal/compiler/lookup/LocalVariableBinding;
    //   255: invokevirtual hasNullInfoFor : (Lorg/eclipse/jdt/internal/compiler/lookup/LocalVariableBinding;)Z
    //   258: ifeq -> 270
    //   261: aload_2
    //   262: aload #7
    //   264: getfield binding : Lorg/eclipse/jdt/internal/compiler/lookup/LocalVariableBinding;
    //   267: invokevirtual markAsDefinitelyNull : (Lorg/eclipse/jdt/internal/compiler/lookup/LocalVariableBinding;)V
    //   270: aload #6
    //   272: aload_2
    //   273: aload_3
    //   274: aload #6
    //   276: aload #4
    //   278: aload #5
    //   280: aload #7
    //   282: invokestatic analyseCloseableExpression : (Lorg/eclipse/jdt/internal/compiler/flow/FlowInfo;Lorg/eclipse/jdt/internal/compiler/flow/FlowContext;Lorg/eclipse/jdt/internal/compiler/lookup/LocalVariableBinding;Lorg/eclipse/jdt/internal/compiler/ast/ASTNode;Lorg/eclipse/jdt/internal/compiler/ast/Expression;Lorg/eclipse/jdt/internal/compiler/ast/FakedTrackingVariable;)Lorg/eclipse/jdt/internal/compiler/ast/FakedTrackingVariable;
    //   285: putfield closeTracker : Lorg/eclipse/jdt/internal/compiler/ast/FakedTrackingVariable;
    //   288: goto -> 337
    //   291: aload_2
    //   292: aload_3
    //   293: aload #6
    //   295: aload #4
    //   297: aload #5
    //   299: aconst_null
    //   300: invokestatic analyseCloseableExpression : (Lorg/eclipse/jdt/internal/compiler/flow/FlowInfo;Lorg/eclipse/jdt/internal/compiler/flow/FlowContext;Lorg/eclipse/jdt/internal/compiler/lookup/LocalVariableBinding;Lorg/eclipse/jdt/internal/compiler/ast/ASTNode;Lorg/eclipse/jdt/internal/compiler/ast/Expression;Lorg/eclipse/jdt/internal/compiler/ast/FakedTrackingVariable;)Lorg/eclipse/jdt/internal/compiler/ast/FakedTrackingVariable;
    //   303: astore #9
    //   305: aload #9
    //   307: ifnull -> 337
    //   310: aload #6
    //   312: aload #9
    //   314: putfield closeTracker : Lorg/eclipse/jdt/internal/compiler/ast/FakedTrackingVariable;
    //   317: aload #9
    //   319: getfield globalClosingState : I
    //   322: bipush #6
    //   324: iand
    //   325: ifne -> 337
    //   328: aload_2
    //   329: aload #9
    //   331: getfield binding : Lorg/eclipse/jdt/internal/compiler/lookup/LocalVariableBinding;
    //   334: invokevirtual markAsDefinitelyNull : (Lorg/eclipse/jdt/internal/compiler/lookup/LocalVariableBinding;)V
    //   337: aload #8
    //   339: ifnull -> 409
    //   342: aload #8
    //   344: getfield innerTracker : Lorg/eclipse/jdt/internal/compiler/ast/FakedTrackingVariable;
    //   347: ifnull -> 383
    //   350: aload #8
    //   352: getfield innerTracker : Lorg/eclipse/jdt/internal/compiler/ast/FakedTrackingVariable;
    //   355: getfield binding : Lorg/eclipse/jdt/internal/compiler/lookup/LocalVariableBinding;
    //   358: getfield declaringScope : Lorg/eclipse/jdt/internal/compiler/lookup/BlockScope;
    //   361: aload_0
    //   362: if_acmpne -> 383
    //   365: aload #8
    //   367: getfield innerTracker : Lorg/eclipse/jdt/internal/compiler/ast/FakedTrackingVariable;
    //   370: aconst_null
    //   371: putfield outerTracker : Lorg/eclipse/jdt/internal/compiler/ast/FakedTrackingVariable;
    //   374: aload_0
    //   375: aload #8
    //   377: invokevirtual pruneWrapperTrackingVar : (Lorg/eclipse/jdt/internal/compiler/ast/FakedTrackingVariable;)V
    //   380: goto -> 409
    //   383: aload_1
    //   384: aload #8
    //   386: getfield binding : Lorg/eclipse/jdt/internal/compiler/lookup/LocalVariableBinding;
    //   389: invokevirtual nullStatus : (Lorg/eclipse/jdt/internal/compiler/lookup/LocalVariableBinding;)I
    //   392: istore #9
    //   394: iload #9
    //   396: iconst_4
    //   397: if_icmpeq -> 409
    //   400: aload #8
    //   402: aload #4
    //   404: iload #9
    //   406: invokevirtual recordErrorLocation : (Lorg/eclipse/jdt/internal/compiler/ast/ASTNode;I)V
    //   409: return
    // Line number table:
    //   Java source line number -> byte code offset
    //   #427	-> 0
    //   #428	-> 3
    //   #429	-> 6
    //   #431	-> 14
    //   #432	-> 21
    //   #433	-> 29
    //   #434	-> 41
    //   #438	-> 45
    //   #440	-> 56
    //   #441	-> 65
    //   #442	-> 70
    //   #444	-> 78
    //   #445	-> 86
    //   #446	-> 93
    //   #449	-> 103
    //   #451	-> 115
    //   #452	-> 118
    //   #453	-> 134
    //   #454	-> 141
    //   #455	-> 142
    //   #456	-> 152
    //   #459	-> 162
    //   #461	-> 181
    //   #464	-> 184
    //   #467	-> 191
    //   #468	-> 199
    //   #470	-> 202
    //   #471	-> 210
    //   #472	-> 213
    //   #476	-> 223
    //   #478	-> 226
    //   #471	-> 233
    //   #482	-> 238
    //   #483	-> 249
    //   #484	-> 261
    //   #485	-> 270
    //   #487	-> 288
    //   #488	-> 291
    //   #489	-> 305
    //   #490	-> 310
    //   #492	-> 317
    //   #493	-> 328
    //   #501	-> 337
    //   #502	-> 342
    //   #504	-> 365
    //   #505	-> 374
    //   #506	-> 380
    //   #507	-> 383
    //   #508	-> 394
    //   #509	-> 400
    //   #512	-> 409
    // Local variable table:
    //   start	length	slot	name	descriptor
    //   0	410	0	scope	Lorg/eclipse/jdt/internal/compiler/lookup/BlockScope;
    //   0	410	1	upstreamInfo	Lorg/eclipse/jdt/internal/compiler/flow/FlowInfo;
    //   0	410	2	flowInfo	Lorg/eclipse/jdt/internal/compiler/flow/FlowInfo;
    //   0	410	3	flowContext	Lorg/eclipse/jdt/internal/compiler/flow/FlowContext;
    //   0	410	4	location	Lorg/eclipse/jdt/internal/compiler/ast/ASTNode;
    //   0	410	5	rhs	Lorg/eclipse/jdt/internal/compiler/ast/Expression;
    //   0	410	6	local	Lorg/eclipse/jdt/internal/compiler/lookup/LocalVariableBinding;
    //   3	407	7	previousTracker	Lorg/eclipse/jdt/internal/compiler/ast/FakedTrackingVariable;
    //   6	404	8	disconnectedTracker	Lorg/eclipse/jdt/internal/compiler/ast/FakedTrackingVariable;
    //   29	16	9	nullStatus	I
    //   65	272	9	rhsTrackVar	Lorg/eclipse/jdt/internal/compiler/ast/FakedTrackingVariable;
    //   202	86	10	currentFlowContext	Lorg/eclipse/jdt/internal/compiler/flow/FlowContext;
    //   394	15	9	upstreamStatus	I
  }
  
  private static org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable analyseCloseableExpression(FlowInfo flowInfo, FlowContext flowContext, LocalVariableBinding local, ASTNode location, Expression expression, org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable previousTracker) {
    while (true) {
      while (expression instanceof Assignment)
        expression = ((Assignment)expression).expression; 
      if (expression instanceof CastExpression) {
        expression = ((CastExpression)expression).expression;
        continue;
      } 
      break;
    } 
    boolean isResourceProducer = false;
    if (expression.resolvedType instanceof ReferenceBinding) {
      ReferenceBinding resourceType = (ReferenceBinding)expression.resolvedType;
      if (resourceType.hasTypeBit(8))
        if (isBlacklistedMethod(expression)) {
          isResourceProducer = true;
        } else {
          return null;
        }  
    } 
    if (expression instanceof AllocationExpression) {
      org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable tracker = ((AllocationExpression)expression).closeTracker;
      if (tracker != null && tracker.originalBinding == null)
        return null; 
      return tracker;
    } 
    if (expression instanceof MessageSend || 
      expression instanceof org.eclipse.jdt.internal.compiler.ast.ArrayReference) {
      org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable tracker = new org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable(local, location, flowInfo, flowContext, 16);
      if (!isResourceProducer)
        tracker.globalClosingState |= 0x2; 
      return tracker;
    } 
    if ((expression.bits & 0x7) == 1 || (
      expression instanceof QualifiedNameReference && (
      (QualifiedNameReference)expression).isFieldAccess())) {
      org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable tracker = new org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable(local, location, flowInfo, flowContext, 1);
      tracker.globalClosingState |= 0x4;
      return tracker;
    } 
    if (local.closeTracker != null)
      return local.closeTracker; 
    org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable newTracker = new org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable(local, location, flowInfo, flowContext, 1);
    LocalVariableBinding rhsLocal = expression.localVariableBinding();
    if (rhsLocal != null && rhsLocal.isParameter())
      newTracker.globalClosingState |= 0x4; 
    return newTracker;
  }
  
  private static boolean isBlacklistedMethod(Expression expression) {
    if (expression instanceof MessageSend) {
      MethodBinding method = ((MessageSend)expression).binding;
      if (method != null && method.isValidBinding())
        return CharOperation.equals(method.declaringClass.compoundName, TypeConstants.JAVA_NIO_FILE_FILES); 
    } 
    return false;
  }
  
  public static void cleanUpAfterAssignment(BlockScope currentScope, int lhsBits, Expression expression) {
    while (true) {
      while (expression instanceof Assignment)
        expression = ((Assignment)expression).expression; 
      if (expression instanceof CastExpression) {
        expression = ((CastExpression)expression).expression;
        continue;
      } 
      break;
    } 
    if (expression instanceof AllocationExpression) {
      org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable tracker = ((AllocationExpression)expression).closeTracker;
      if (tracker != null && tracker.originalBinding == null) {
        tracker.withdraw();
        ((AllocationExpression)expression).closeTracker = null;
      } 
    } else {
      LocalVariableBinding local = expression.localVariableBinding();
      if (local != null && local.closeTracker != null && (lhsBits & 0x1) != 0)
        local.closeTracker.withdraw(); 
    } 
  }
  
  public static boolean isAnyCloseable(TypeBinding typeBinding) {
    return (typeBinding instanceof ReferenceBinding && (
      (ReferenceBinding)typeBinding).hasTypeBit(3));
  }
  
  public int findMostSpecificStatus(FlowInfo flowInfo, BlockScope currentScope, BlockScope locationScope) {
    int status = 1;
    org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable currentTracker = this;
    while (currentTracker != null) {
      LocalVariableBinding currentVar = currentTracker.binding;
      int currentStatus = getNullStatusAggressively(currentVar, flowInfo);
      if (locationScope != null)
        currentStatus = mergeCloseStatus(locationScope, currentStatus, currentVar, currentScope); 
      if (currentStatus == 4) {
        status = currentStatus;
        break;
      } 
      if (status == 2 || status == 1)
        status = currentStatus; 
      currentTracker = currentTracker.innerTracker;
    } 
    return status;
  }
  
  private int getNullStatusAggressively(LocalVariableBinding local, FlowInfo flowInfo) {
    if (flowInfo == FlowInfo.DEAD_END)
      return 1; 
    int reachMode = flowInfo.reachMode();
    int status = 0;
    try {
      if (reachMode != 0)
        flowInfo.tagBits &= 0xFFFFFFFC; 
      status = flowInfo.nullStatus(local);
      if (TEST_372319)
        try {
          Thread.sleep(5L);
        } catch (InterruptedException interruptedException) {} 
    } finally {
      flowInfo.tagBits |= reachMode;
    } 
    if ((status & 0x2) != 0) {
      if ((status & 0x24) != 0)
        return 16; 
      return 2;
    } 
    if ((status & 0x4) != 0) {
      if ((status & 0x10) != 0)
        return 16; 
      return 4;
    } 
    if ((status & 0x10) != 0)
      return 16; 
    return status;
  }
  
  public int mergeCloseStatus(BlockScope currentScope, int status, LocalVariableBinding local, BlockScope outerScope) {
    if (status != 4) {
      if (currentScope.finallyInfo != null) {
        int finallyStatus = currentScope.finallyInfo.nullStatus(local);
        if (finallyStatus == 4)
          return finallyStatus; 
        if (finallyStatus != 2 && currentScope.finallyInfo.hasNullInfoFor(local))
          status = 16; 
      } 
      if (currentScope != outerScope && currentScope.parent instanceof BlockScope)
        return mergeCloseStatus((BlockScope)currentScope.parent, status, local, outerScope); 
    } 
    return status;
  }
  
  public void markClose(FlowInfo flowInfo, FlowContext flowContext) {
    org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable current = this;
    do {
      flowInfo.markAsDefinitelyNonNull(current.binding);
      current.globalClosingState |= 0x1;
      flowContext.markFinallyNullStatus(current.binding, 4);
      current = current.innerTracker;
    } while (current != null);
  }
  
  public void markClosedInNestedMethod() {
    this.globalClosingState |= 0x8;
  }
  
  public static FlowInfo markPassedToOutside(BlockScope scope, Expression expression, FlowInfo flowInfo, FlowContext flowContext, boolean owned) {
    org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable trackVar = getCloseTrackingVariable(expression, flowInfo, flowContext);
    if (trackVar != null) {
      FlowInfo infoResourceIsClosed = owned ? flowInfo : flowInfo.copy();
      int flag = owned ? 4 : 2;
      while (true) {
        trackVar.globalClosingState |= flag;
        if (scope.methodScope() != trackVar.methodScope)
          trackVar.globalClosingState |= 0x8; 
        infoResourceIsClosed.markAsDefinitelyNonNull(trackVar.binding);
        if ((trackVar = trackVar.innerTracker) == null) {
          if (owned)
            return infoResourceIsClosed; 
          return FlowInfo.conditional(flowInfo, infoResourceIsClosed);
        } 
      } 
    } 
    return flowInfo;
  }
  
  public boolean hasDefinitelyNoResource(FlowInfo flowInfo) {
    if (this.originalBinding == null)
      return false; 
    if (flowInfo.isDefinitelyNull(this.originalBinding))
      return true; 
    if (!flowInfo.isDefinitelyAssigned(this.originalBinding) && 
      !flowInfo.isPotentiallyAssigned(this.originalBinding))
      return true; 
    return false;
  }
  
  public boolean isClosedInFinallyOfEnclosing(BlockScope scope) {
    BlockScope currentScope = scope;
    while (true) {
      if (currentScope.finallyInfo != null && 
        currentScope.finallyInfo.isDefinitelyNonNull(this.binding))
        return true; 
      if (!(currentScope.parent instanceof BlockScope))
        return false; 
      currentScope = (BlockScope)currentScope.parent;
    } 
  }
  
  public boolean isResourceBeingReturned(org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable returnedResource) {
    org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable current = this;
    while (true) {
      if (current == returnedResource) {
        this.globalClosingState |= 0x40;
        return true;
      } 
      current = current.innerTracker;
      if (current == null)
        return false; 
    } 
  }
  
  public void withdraw() {
    this.binding.declaringScope.removeTrackingVar(this);
  }
  
  public void recordErrorLocation(ASTNode location, int nullStatus) {
    if ((this.globalClosingState & 0x4) != 0)
      return; 
    if (this.recordedLocations == null)
      this.recordedLocations = new HashMap<>(); 
    this.recordedLocations.put(location, Integer.valueOf(nullStatus));
  }
  
  public boolean reportRecordedErrors(Scope scope, int mergedStatus, boolean atDeadEnd) {
    org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable current = this;
    while (current.globalClosingState == 0) {
      current = current.innerTracker;
      if (current == null) {
        if (atDeadEnd && neverClosedAtLocations())
          mergedStatus = 2; 
        if ((mergedStatus & 0x32) != 0) {
          reportError(scope.problemReporter(), null, mergedStatus);
          return true;
        } 
        break;
      } 
    } 
    boolean hasReported = false;
    if (this.recordedLocations != null) {
      Iterator<Map.Entry> locations = this.recordedLocations.entrySet().iterator();
      int reportFlags = 0;
      while (locations.hasNext()) {
        Map.Entry entry = locations.next();
        reportFlags |= reportError(scope.problemReporter(), (ASTNode)entry.getKey(), ((Integer)entry.getValue()).intValue());
        hasReported = true;
      } 
      if (reportFlags != 0) {
        current = this;
        do {
          current.globalClosingState |= reportFlags;
        } while ((current = current.innerTracker) != null);
      } 
    } 
    return hasReported;
  }
  
  private boolean neverClosedAtLocations() {
    if (this.recordedLocations != null)
      for (Object value : this.recordedLocations.values()) {
        if (!value.equals(Integer.valueOf(2)))
          return false; 
      }  
    return true;
  }
  
  public int reportError(ProblemReporter problemReporter, ASTNode location, int nullStatus) {
    if ((this.globalClosingState & 0x4) != 0)
      return 0; 
    boolean isPotentialProblem = false;
    if (nullStatus == 2) {
      if ((this.globalClosingState & 0x8) != 0)
        isPotentialProblem = true; 
    } else if ((nullStatus & 0x30) != 0) {
      isPotentialProblem = true;
    } 
    if (isPotentialProblem) {
      if ((this.globalClosingState & 0x60) != 0)
        return 0; 
      problemReporter.potentiallyUnclosedCloseable(this, location);
    } else {
      if ((this.globalClosingState & 0x40) != 0)
        return 0; 
      problemReporter.unclosedCloseable(this, location);
    } 
    int reportFlag = isPotentialProblem ? 32 : 64;
    if (location == null) {
      org.eclipse.jdt.internal.compiler.ast.FakedTrackingVariable current = this;
      do {
        current.globalClosingState |= reportFlag;
      } while ((current = current.innerTracker) != null);
    } 
    return reportFlag;
  }
  
  public void reportExplicitClosing(ProblemReporter problemReporter) {
    if ((this.globalClosingState & 0x14) == 0) {
      this.globalClosingState |= 0x10;
      problemReporter.explicitlyClosedAutoCloseable(this);
    } 
  }
  
  public String nameForReporting(ASTNode location, ReferenceContext referenceContext) {
    if (this.name == UNASSIGNED_CLOSEABLE_NAME && 
      location != null && referenceContext != null) {
      CompilationResult compResult = referenceContext.compilationResult();
      if (compResult != null) {
        int[] lineEnds = compResult.getLineSeparatorPositions();
        int resourceLine = Util.getLineNumber(this.sourceStart, lineEnds, 0, lineEnds.length - 1);
        int reportLine = Util.getLineNumber(location.sourceStart, lineEnds, 0, lineEnds.length - 1);
        if (resourceLine != reportLine) {
          char[] replacement = Integer.toString(resourceLine).toCharArray();
          return String.valueOf(CharOperation.replace(UNASSIGNED_CLOSEABLE_NAME_TEMPLATE, TEMPLATE_ARGUMENT, replacement));
        } 
      } 
    } 
    return String.valueOf(this.name);
  }
}
